﻿using DEVSsharp;
using Stefan_Stamenkovic_91_2014.Other;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Stefan_Stamenkovic_91_2014.Models
{
    class Porcupine : Atomic
    {
        public Vector Position { get; set; }
        public double T { get; set; }

        public InputPort InputTemp { get; }
        public InputPort InputWait { get; }
        public InputPort InputGoTo { get; }

        public OutputPort OutputTemp { get; }
        public OutputPort OutputPosition { get; }

        private Triangle currentTriangle;
        private double Kv;
        private double Vmax;
        private double Vmin;
        private double epsD;
        private double K0;
        private double Kk;
        private double Ks;
        private double epsT;
        private double Ts;

        private int waitFor = 0;

        private List<PortValue> buffer;
        private Vector goTo;
        private NextChange nextChange;
        private Vector direction;

        private List<TauValue> tauBuffer;

        public Porcupine(string name, TimeUnit tu, Triangle currentTriangle, Vector position, double Kv, double Vmax, double Vmin, double epsD, double T, double Ts, double K0, double Kk, double Ks, double epsT) : base(name, tu)
        {
            this.Kv = Kv;
            this.Vmax = Vmax;
            this.Vmin = Vmin;
            this.epsD = epsD;
            this.T = T;
            this.Ts = Ts;
            this.K0 = K0;
            this.Kk = Kk;
            this.Ks = Ks;
            this.epsT = epsT;
            this.currentTriangle = currentTriangle;
            this.Position = position;

            InputTemp = AddIP("InputTemp" + name);
            InputGoTo = AddIP("InputGoTo" + name);
            InputWait = AddIP("InputWait" + name);

            OutputTemp = AddOP("OutputTemp" + name);
            OutputPosition = AddOP("OutputPosition" + name);
        }

        public override bool delta_x(PortValue x)
        {
            if(InputWait==x.port)
            {
                bool wait = (bool)x.value;

                if(wait)
                {
                    goTo = null;
                    CalculateDirection();
                    waitFor++;
                }
                else
                {
                    waitFor--;
                }
            }
            else if(InputGoTo==x.port)
            {
                goTo = x.value as Vector;
                CalculateDirection();
            }
            else
            {
                nextChange = NextChange.TEMP;
                RemoveTempTau();
                CalculateChanges();
                T = (double) x.value;
            }

            return true;
        }

        public override void delta_y(ref PortValue y)
        {
            if(buffer.Any())
            {
                y = buffer[0];
                buffer.RemoveAt(0);
            }
            else
            {
                CalculateChanges();
            }
        }

        public override void init()
        {
            buffer = new List<PortValue>();
            tauBuffer = new List<TauValue>();
            CalculateDirection();

            tauBuffer.Add(new TauValue()
            {
                Time = CalculateMoveTau(),
                NextChange = NextChange.POSITION
            });

            AddSortedTau(new TauValue()
            {
                Time = CalculateTempTau(),
                NextChange = NextChange.TEMP
            });
        }

        public override double tau()
        {
            if(buffer.Any())
            {
                return 0;
            }

            TauValue tv = tauBuffer[0];
            tauBuffer.RemoveAt(0);

            nextChange = tv.NextChange;
            return tv.Time - TimeCurrent;
        }

        private void CalculateChanges()
        {
            if (nextChange == NextChange.POSITION)
            {
                Position += direction * TimeElapsed;
                IsOutCheck();

                AddSortedTau(new TauValue()
                {
                    Time = TimeCurrent + CalculateMoveTau(),
                    NextChange = NextChange.POSITION
                });
            }
            else if (nextChange == NextChange.TEMP) 
            {
                T = (K0+Kk*direction.Norm()+Ks*T-Math.Pow(Math.E, TimeElapsed)) / Ks;

                AddSortedTau(new TauValue()
                {
                    Time = TimeCurrent + CalculateTempTau(),
                    NextChange = NextChange.TEMP
                });
            }
        }

        private double CalculateMoveTau()
        {
            return epsD / direction.Norm();
        }

        private double CalculateTempTau()
        {
            double norm = direction.Norm();
            int sign = Math.Sign(Ts - T);

            return Math.Log(K0 + Kk * norm + Ks * epsT * sign);
        }

        private void CalculateDirection()
        {
            Vector normalOnXY = new Vector(0, 0, 1);

            if (goTo==null)
            {
                Vector normalOnTriangle = GetNormalVector(currentTriangle.A, currentTriangle.B, currentTriangle.C);

                direction = normalOnTriangle.CrossProduct(normalOnXY).CrossProduct(normalOnTriangle);
            }
            else
            {
                Vector rotateAround = normalOnXY.CrossProduct(goTo);
                direction = goTo.Rotate(rotateAround, Position, GetAngle());
            }

            SetSpeed();
        }

        private void IsOutCheck()
        {
            if(!OnSameSide(Position, currentTriangle.A, currentTriangle.B, currentTriangle.C))
            {
                if(currentTriangle.SideA!=null)
                {
                    ReflectDirection(currentTriangle.A, currentTriangle.B, currentTriangle.C);
                }
                else
                {
                    currentTriangle = currentTriangle.SideA;
                    CalculateDirection();
                }
            }
            else if(!OnSameSide(Position, currentTriangle.B, currentTriangle.A, currentTriangle.C))
            {
                if (currentTriangle.SideB != null)
                {
                    ReflectDirection(currentTriangle.B, currentTriangle.A, currentTriangle.C);
                }
                else
                {
                    currentTriangle = currentTriangle.SideB;
                    CalculateDirection();
                }
            }
            else if(!OnSameSide(Position, currentTriangle.C, currentTriangle.A, currentTriangle.B))
            {

                if (currentTriangle.SideC != null)
                {
                    ReflectDirection(currentTriangle.C, currentTriangle.A, currentTriangle.B);
                }
                else
                {
                    currentTriangle = currentTriangle.SideC;
                    CalculateDirection();
                }
            }
        }

        private bool OnSameSide(Vector p1, Vector p2, Vector A, Vector B)
        {
            double d1 = (p1.X - A.X) * (B.Y - A.Y) - (p1.Y - A.Y) * (B.X - A.X);
            double d2 = (p2.X - A.X) * (B.Y - A.Y) - (p2.Y - A.Y) * (B.X - A.X);

            return Math.Sign(d1) == Math.Sign(d2);
        }

        private void ReflectDirection(Vector to, Vector A, Vector B)
        {
            Vector normal = GetNormalVector(to, A, B);

            direction = normal.CrossProduct(A-B);

            SetSpeed();
        }

        private Vector GetNormalVector(Vector A, Vector B, Vector C)
        {
            return (B - C).CrossProduct(B - A);
        }

        private void SetSpeed()
        {
            double angle = GetAngle();

            direction *= ClampSpeed(angle) / direction.Norm();
        }

        private double ClampSpeed(double angle)
        {
            double speed = angle * Kv;

            if (speed < Vmin)
            {
                speed = Vmin;
            }
            else if (speed > Vmax)
            {
                speed = Vmax;
            }

            return speed;
        }

        private void AddSortedTau(TauValue tv)
        {
            int i;

            for (i = 0; i < tauBuffer.Count && tauBuffer[i].Time<tv.Time; i++);

            tauBuffer.Insert(i, tv);
        }

        private void RemoveTempTau()
        {
            tauBuffer.Remove(tauBuffer.FirstOrDefault(x=>x.NextChange == NextChange.TEMP));
        }

        private double GetAngle()
        {
            Vector bottom = direction.Clone();
            bottom.Z = 0;

            double cosFi = (direction * bottom) / (direction.Norm() * bottom.Norm());
            double sinFi = direction.CrossProduct(bottom).Norm() / (direction.Norm() * bottom.Norm());

            return Math.Atan2(sinFi, cosFi);
        }

        public override string Get_s()
        {
            return String.Format("Positon: {0}, Direction: {1}", Position, direction);
        }
    }
}
